from typing import List, Mapping, Union
from warnings import warn

from citrine._rest.engine_resource import EngineResource
from citrine._serialization import properties
from citrine.informatics.descriptors import Descriptor
from citrine.informatics.design_spaces.design_space import DesignSpace

__all__ = ['EnumeratedDesignSpace']


class EnumeratedDesignSpace(EngineResource['EnumeratedDesignSpace'], DesignSpace):
    """An explicit enumeration of candidate materials to score.

    Enumerated design spaces are intended to capture small spaces with fewer than
    1000 values.  For larger spaces, use the DataSourceDesignSpace.

    Parameters
    ----------
    name:str
        the name of the design space
    description:str
        the description of the design space
    descriptors: list[Descriptor]
        the list of descriptors included in the candidates of the design space
    data: list[dict]
        list of dicts of the shape `{<descriptor_key>: <descriptor_value>}`
        where each dict corresponds to a candidate in the design space

    """

    descriptors = properties.List(properties.Object(Descriptor), 'data.instance.descriptors')
    _data = properties.List(properties.Mapping(properties.String,
                                               properties.Union([properties.String(),
                                                                 properties.Integer(),
                                                                 properties.Float()])),
                            'data.instance.data')

    typ = properties.String('data.instance.type', default='EnumeratedDesignSpace',
                            deserializable=False)

    def __init__(self,
                 name: str,
                 *,
                 description: str,
                 descriptors: List[Descriptor],
                 data: List[Mapping[str, Union[int, float, str]]]):
        self.name: str = name
        self.description: str = description
        self.descriptors: List[Descriptor] = descriptors
        self.data: List[Mapping[str, Union[int, float, str]]] = data

    def __str__(self):
        return '<EnumeratedDesignSpace {!r}>'.format(self.name)

    @property
    def data(self) -> List[Mapping[str, Union[int, float, str]]]:
        """List of dicts corresponding to candidates in the design space."""
        return self._data

    @data.setter
    def data(self, value: List[Mapping[str, Union[int, float, str]]]):
        for item in value:
            for el in item.values():
                if isinstance(el, (int, float)):
                    warn("Providing numeric data values is deprecated as of 3.4.7, and will be "
                         "dropped in 4.0.0. Please use strings instead.",
                         DeprecationWarning)
        self._data = value



from typing import List, Optional, Union
from uuid import UUID

from citrine._rest.engine_resource import EngineResourceWithoutStatus
from citrine._rest.resource import Resource
from citrine._serialization import properties
from citrine.informatics.workflows.workflow import Workflow
from citrine.gemd_queries.gemd_query import GemdQuery


class LatestBuild(Resource['LatestBuild']):
    """Info on the latest analysis workflow build."""

    status = properties.Optional(properties.String, 'status', serializable=False)
    failures = properties.List(properties.String, 'failure_reason', default=[], serializable=False)
    query = properties.Optional(properties.Object(GemdQuery), 'query', serializable=False)


class AnalysisWorkflow(EngineResourceWithoutStatus['AnalysisWorkflow'], Workflow):
    """An analysis workflow stored on the platform.

    Note that plots are not fully supported. They're captured as raw JSON in order to facilitate
    cloning an existing workflow, but no facilities are provided to validate them in the client.
    """

    uid = properties.UUID('id', serializable=False)
    name = properties.String('data.name')
    description = properties.String('data.description')
    snapshot_id = properties.Optional(properties.UUID, 'data.snapshot_id')
    _plots = properties.List(properties.Raw, 'data.plots', default=[])

    latest_build = properties.Optional(properties.Object(LatestBuild), 'metadata.latest_build',
                                       serializable=False)

    def __init__(self,
                 *,
                 name: str,
                 description: str,
                 snapshot_id: Optional[Union[UUID, str]] = None,
                 plots: List[dict] = []):
        self.name = name
        self.description = description
        self.snapshot_id = snapshot_id
        self._plots = plots

    @property
    def status(self) -> str:
        """An alias for 'latest_build.status', to allow its usage with 'wait_while_validating."""
        return self.latest_build.status

    def _post_dump(self, data: dict) -> dict:
        data["data"]["plots"] = [plot["data"] for plot in data["data"].get("plots") or []]
        return super()._post_dump(data)


class AnalysisWorkflowUpdatePayload(Resource['AnalysisWorkflowUpdatePayload']):
    """An object capturing the analysis workflow upload payload.

    Making this a separate payload makes it explicit that you can only update name and description.
    Additionally, unlike in AnalysisWorkflow, both are optional, to allow changing one without
    changing the other.
    """

    uid = properties.UUID('id', serializable=False)
    name = properties.Optional(properties.String, 'name')
    description = properties.Optional(properties.String, 'description')

    def __init__(self,
                 uid: Union[UUID, str],
                 *,
                 name: Optional[str] = None,
                 description: Optional[str] = None):
        self.uid = uid
        self.name = name
        self.description = description



"""Resources that represent both individual and collections of teams."""
from typing import List, Optional, Tuple, Union
from uuid import UUID

from gemd.entity.base_entity import BaseEntity
from gemd.entity.link_by_uid import LinkByUID

from citrine._rest.admin_collection import AdminCollection
from citrine._rest.resource import Resource, ResourceTypeEnum
from citrine._serialization import properties
from citrine._session import Session
from citrine._utils.functions import format_escaped_url
from citrine.resources.analysis_workflow import AnalysisWorkflowCollection
from citrine.resources.api_error import ApiError
from citrine.resources.condition_template import ConditionTemplateCollection
from citrine.resources.dataset import DatasetCollection
from citrine.resources.delete import _async_gemd_batch_delete
from citrine.resources.gemd_resource import GEMDResourceCollection
from citrine.resources.ingredient_run import IngredientRunCollection
from citrine.resources.ingredient_spec import IngredientSpecCollection
from citrine.resources.material_run import MaterialRunCollection
from citrine.resources.material_spec import MaterialSpecCollection
from citrine.resources.material_template import MaterialTemplateCollection
from citrine.resources.measurement_run import MeasurementRunCollection
from citrine.resources.measurement_spec import MeasurementSpecCollection
from citrine.resources.measurement_template import MeasurementTemplateCollection
from citrine.resources.parameter_template import ParameterTemplateCollection
from citrine.resources.process_run import ProcessRunCollection
from citrine.resources.process_spec import ProcessSpecCollection
from citrine.resources.process_template import ProcessTemplateCollection
from citrine.resources.project import ProjectCollection
from citrine.resources.property_template import PropertyTemplateCollection
from citrine.resources.user import User, UserCollection


WRITE = "WRITE"
READ = "READ"
SHARE = "SHARE"
TEAM_ACTIONS = Union[WRITE, READ, SHARE]


class TeamMember:
    """A Member of a Team."""

    def __init__(self,
                 *,
                 user: User,
                 team: 'Team',  # noqa: F821
                 actions: List[TEAM_ACTIONS]):
        self.user = user
        self.team: 'Team' = team  # noqa: F821
        self.actions: List[TEAM_ACTIONS] = actions

    def __str__(self):
        return '<TeamMember {!r} can {!s} in {!r}>' \
            .format(self.user.screen_name, self.actions, self.team.name)


class TeamResourceIDs:
    """
    A Citrine Team Resource IDs class.

    This class is used to list the unique IDs for specific resource
    types published in a single team and therefore available to be pulled
    in by all projects.

    Parameters
    ----------
    team_id: str or uuid
        ID of the team.
    resource_type: str
        The resource type to list, one of DATASET/MODULE/TABLE/TABLE_DEFINITION

    """

    _api_version = "v3"

    def __init__(self,
                 session: Session,
                 team_id: Union[str, UUID],
                 resource_type: str) -> None:
        self.session = session
        self.team_id = team_id
        self.resource_type = resource_type

    def _path(self) -> str:
        return format_escaped_url(f'/teams/{self.team_id}')

    def _list_ids(self, action: str) -> List[str]:
        query_params = {"domain": self._path(), "action": action}
        return self.session.get_resource(f"/{self.resource_type}/authorized-ids",
                                         params=query_params,
                                         version=self._api_version)['ids']

    def list_readable(self):
        """
        List IDs of the published resources with READ access.

        Returns
        -------
        List[str]
            Returns a list of string UUIDs for each resource

        """
        return self._list_ids(action=READ)

    def list_writeable(self):
        """
        List IDs of the published resources with WRITE access.

        Returns
        -------
        List[str]
            Returns a list of string UUIDs for each resource

        """
        return self._list_ids(action=WRITE)

    def list_shareable(self):
        """
        List IDs of the published resources with SHARE access.

        Returns
        -------
        List[str]
            Returns a list of string UUIDs for each resource

        """
        return self._list_ids(action=SHARE)


class Team(Resource['Team']):
    """
    A Citrine Team.

    A Team is a collection of projects and resources.

    Parameters
    ----------
    name: str
        Name of the team.
    description: str
        Long-form description of the team.
    session: Session, optional
        The Citrine session used to connect to the database.

    """

    _response_key = 'team'
    _resource_type = ResourceTypeEnum.TEAM
    _api_version = "v3"

    name = properties.String('name')
    """str: Name of the Team"""
    description = properties.Optional(properties.String(), 'description')
    """str: Description of the Team"""
    uid = properties.Optional(properties.UUID(), 'id')
    """UUID: Unique uuid4 identifier of this team."""
    created_at = properties.Optional(properties.Datetime(), 'created_at')
    """int: Time the team was created, in seconds since epoch."""

    def __init__(self,
                 name: str,
                 *,
                 description: str = "",
                 session: Optional[Session] = None):
        self.name: str = name
        self.description: str = description
        self.session: Session = session

    def __str__(self):
        return '<Team {!r}>'.format(self.name)

    def _path(self):
        return format_escaped_url('/teams/{team_id}', team_id=self.uid)

    def list_members(self) -> List[TeamMember]:
        """
        List all of the members in the current team.

        Requires admin privileges.

        Returns
        -------
        List[TeamMember]
            The members of the current team

        """
        response = self.session.get_resource(self._path() + "/users", version=self._api_version)
        members = response["users"]
        return [TeamMember(user=User.build(m), team=self, actions=m["actions"]) for m in members]

    def get_member(self, user_id: Union[str, UUID, User]) -> TeamMember:
        """
        Get a particular member in the current team.

        May require admin privileges depending on which user is being requested.

        Parameters
        ----------
        user_id: str or uuid
            The id of the user to remove from the team

        Returns
        -------
        TeamMember
            The requested team member

        """
        if isinstance(user_id, User):
            user_id = user_id.uid
        path = self._path() + format_escaped_url('/users/{user_id}', user_id=user_id)
        member = self.session.get_resource(path=path, version=self._api_version)["user"]
        return TeamMember(user=User.build(member), team=self, actions=member["actions"])

    def me(self) -> TeamMember:
        """
        Get the member for the current user.

        Returns
        -------
        TeamMember
            The TeamMember object representing the current user

        """
        me = UserCollection(self.session).me()
        return self.get_member(me)

    def remove_user(self, user_id: Union[str, UUID, User]) -> bool:
        """
        Remove a User from a Team.

        Requires admin privileges.

        Parameters
        ----------
        user_id: User, str or uuid
            The id of the user to remove from the team

        Returns
        -------
        bool
            Returns ``True`` if user successfully removed

        """
        if isinstance(user_id, User):
            user_id = user_id.uid
        self.session.checked_post(self._path() + "/users/batch-remove",
                                  json={"ids": [str(user_id)]}, version=self._api_version)
        return True  # note: only get here if checked_post doesn't raise error

    def add_user(self,
                 user_id: Union[str, UUID, User],
                 *,
                 actions: Optional[List[TEAM_ACTIONS]] = None) -> bool:
        """
        Add a User to a Team.

        Requires admin privileges.

        If no actions are specified, adds User with ``READ`` action to the Team.

        Use the `update_user_action` method to change a User's actions.

        Parameters
        ----------
        user_id: User, str or uuid
            The id of the user to add to the team
        actions: list of TEAM_ACTIONS
            The actions to give the new user in this team
            The options are: WRITE, READ, SHARE
            defaults to READ

        Returns
        -------
        bool
            Returns ``True`` if user successfully added

        """
        if isinstance(user_id, User):
            user_id = user_id.uid
        if actions is None:
            actions = [READ]
        return self.update_user_action(user_id, actions=actions)

    def update_user_action(self,
                           user_id: Union[str, UUID, User],
                           *,
                           actions: List[TEAM_ACTIONS]) -> bool:
        """
        Overwrites a User's action permissions in the Team.

        Requires admin privileges.

        Parameters
        ----------
        user_id: User, str or uuid
            The id of the user to add to the team
        actions: list of TEAM_ACTIONS
            The actions to give the new user in this team
            The options are: WRITE, READ, SHARE

        Returns
        -------
        bool
            Returns ``True`` if user successfully added

        """
        if isinstance(user_id, User):
            user_id = user_id.uid
        self.session.checked_put(self._path() + "/users", version=self._api_version,
                                 json={'id': str(user_id), "actions": actions})
        return True

    def share(self,
              *,
              resource: Resource,
              target_team_id: Union[str, UUID, "Team"]) -> bool:
        """
        Share a resource with another team.

        Requires SHARE action.

        Parameters
        ----------
        resource: Resource
            The resource owned by this team, which will be shared
        target_team_id: Union[str, UUID, Team]
            The id of the team with which to share the resource

        Returns
        -------
        bool
            Returns ``True`` if resource successfully shared

        """
        if isinstance(target_team_id, Team):
            target_team_id = target_team_id.uid
        resource_access = resource.access_control_dict()
        payload = {
            "resource_type": resource_access["type"],
            "resource_id": resource_access["id"],
            "target_team_id": str(target_team_id)
        }
        self.session.checked_post(self._path() + "/shared-resources",
                                  version=self._api_version, json=payload)
        return True

    def un_share(self, *, resource: Resource, target_team_id: Union[str, UUID, "Team"]) -> bool:
        """
        Revoke the share of a particular resource to a secondary team.

        Requires SHARE action.

        Parameters
        ----------
        resource: Resource
            The resource owned by this team, which will be un-shared
        target_team_id: Union[str, UUID, Team]
            The id of the team which should not have access to the resource

        Returns
        -------
        bool
            Returns ``True`` if resource successfully un-shared

        """
        if isinstance(target_team_id, Team):
            target_team_id = target_team_id.uid
        resource_type = resource.access_control_dict()["type"]
        resource_id = resource.access_control_dict()["id"]
        self.session.checked_delete(
            self._path() + f"/shared-resources/{resource_type}/{resource_id}",
            version=self._api_version,
            json={"target_team_id": str(target_team_id)}
        )
        return True

    def owned_dataset_ids(self) -> List[str]:
        """
        List all the ids of the datasets owned by the current team.

        Returns
        -------
        List[str]
            The ids of the modules owned by current team

        """
        query_params = {"userId": "", "domain": self._path(), "action": "WRITE"}
        response = self.session.get_resource("/DATASET/authorized-ids",
                                             params=query_params,
                                             version="v3")
        return response['ids']

    @property
    def projects(self) -> ProjectCollection:
        """Return a resource representing all visible projects in this team."""
        return ProjectCollection(self.session, team_id=self.uid)

    @property
    def analyses(self) -> AnalysisWorkflowCollection:
        """Return a resource representing all visible analysis workflows in this team."""
        return AnalysisWorkflowCollection(session=self.session, team_id=self.uid)

    @property
    def dataset_ids(self) -> TeamResourceIDs:
        """Return a TeamResourceIDs instance for listing published dataset IDs."""
        return TeamResourceIDs(session=self.session,
                               team_id=self.uid,
                               resource_type=ResourceTypeEnum.DATASET.value)

    @property
    def datasets(self) -> DatasetCollection:
        """Return a resource representing all visible datasets."""
        return DatasetCollection(team_id=self.uid, session=self.session)

    @property
    def module_ids(self) -> TeamResourceIDs:
        """Return a TeamResourceIDs instance for listing published module IDs."""
        return TeamResourceIDs(session=self.session,
                               team_id=self.uid,
                               resource_type=ResourceTypeEnum.MODULE.value)

    @property
    def table_ids(self) -> TeamResourceIDs:
        """Return a TeamResourceIDs instance for listing published table IDs."""
        return TeamResourceIDs(session=self.session,
                               team_id=self.uid,
                               resource_type=ResourceTypeEnum.TABLE.value)

    @property
    def table_definition_ids(self) -> TeamResourceIDs:
        """Return a TeamResourceIDs instance for listing published table definition IDs."""
        return TeamResourceIDs(session=self.session,
                               team_id=self.uid,
                               resource_type=ResourceTypeEnum.TABLE_DEFINITION.value)

    @property
    def property_templates(self) -> PropertyTemplateCollection:
        """Return a resource representing all property templates in this dataset."""
        return PropertyTemplateCollection(team_id=self.uid, dataset_id=None, session=self.session)

    @property
    def condition_templates(self) -> ConditionTemplateCollection:
        """Return a resource representing all condition templates in this dataset."""
        return ConditionTemplateCollection(team_id=self.uid, dataset_id=None, session=self.session)

    @property
    def parameter_templates(self) -> ParameterTemplateCollection:
        """Return a resource representing all parameter templates in this dataset."""
        return ParameterTemplateCollection(team_id=self.uid, dataset_id=None, session=self.session)

    @property
    def material_templates(self) -> MaterialTemplateCollection:
        """Return a resource representing all material templates in this dataset."""
        return MaterialTemplateCollection(team_id=self.uid, dataset_id=None, session=self.session)

    @property
    def measurement_templates(self) -> MeasurementTemplateCollection:
        """Return a resource representing all measurement templates in this dataset."""
        return MeasurementTemplateCollection(team_id=self.uid,
                                             dataset_id=None,
                                             session=self.session)

    @property
    def process_templates(self) -> ProcessTemplateCollection:
        """Return a resource representing all process templates in this dataset."""
        return ProcessTemplateCollection(team_id=self.uid, dataset_id=None, session=self.session)

    @property
    def process_runs(self) -> ProcessRunCollection:
        """Return a resource representing all process runs in this dataset."""
        return ProcessRunCollection(team_id=self.uid, dataset_id=None, session=self.session)

    @property
    def measurement_runs(self) -> MeasurementRunCollection:
        """Return a resource representing all measurement runs in this dataset."""
        return MeasurementRunCollection(team_id=self.uid, dataset_id=None, session=self.session)

    @property
    def material_runs(self) -> MaterialRunCollection:
        """Return a resource representing all material runs in this dataset."""
        return MaterialRunCollection(team_id=self.uid, dataset_id=None, session=self.session)

    @property
    def ingredient_runs(self) -> IngredientRunCollection:
        """Return a resource representing all ingredient runs in this dataset."""
        return IngredientRunCollection(team_id=self.uid, dataset_id=None, session=self.session)

    @property
    def process_specs(self) -> ProcessSpecCollection:
        """Return a resource representing all process specs in this dataset."""
        return ProcessSpecCollection(team_id=self.uid, dataset_id=None, session=self.session)

    @property
    def measurement_specs(self) -> MeasurementSpecCollection:
        """Return a resource representing all measurement specs in this dataset."""
        return MeasurementSpecCollection(team_id=self.uid, dataset_id=None, session=self.session)

    @property
    def material_specs(self) -> MaterialSpecCollection:
        """Return a resource representing all material specs in this dataset."""
        return MaterialSpecCollection(team_id=self.uid, dataset_id=None, session=self.session)

    @property
    def ingredient_specs(self) -> IngredientSpecCollection:
        """Return a resource representing all ingredient specs in this dataset."""
        return IngredientSpecCollection(team_id=self.uid, dataset_id=None, session=self.session)

    @property
    def gemd(self) -> GEMDResourceCollection:
        """Return a resource representing all GEMD objects/templates in this dataset."""
        return GEMDResourceCollection(team_id=self.uid, dataset_id=None, session=self.session)

    def gemd_batch_delete(self,
                          id_list: List[Union[LinkByUID, UUID, str, BaseEntity]],
                          *,
                          timeout: float = 2 * 60,
                          polling_delay: float = 1.0) -> List[Tuple[LinkByUID, ApiError]]:
        """
        Remove a set of GEMD objects.

        You may provide GEMD objects that reference each other, and the objects
        will be removed in the appropriate order.

        A failure will be returned if the object cannot be deleted due to an external
        reference.

        You must have Write access on the associated datasets for each object.

        Parameters
        ----------
        id_list: List[Union[LinkByUID, UUID, str, BaseEntity]]
            A list of the IDs of data objects to be removed. They can be passed
            as a LinkByUID tuple, a UUID, a string, or the object itself. A UUID
            or string is assumed to be a Citrine ID, whereas a LinkByUID or
            BaseEntity can also be used to provide an external ID.
        timeout: float
            Amount of time to wait on the job (in seconds) before giving up. Defaults
            to 2 minutes. Note that this number has no effect on the underlying job
            itself, which can also time out server-side.
        polling_delay: float
            How long to delay between each polling retry attempt (in seconds).

        Returns
        -------
        List[Tuple[LinkByUID, ApiError]]
            A list of (LinkByUID, api_error) for each failure to delete an object.
            Note that this method doesn't raise an exception if an object fails to be
            deleted.

        """
        return _async_gemd_batch_delete(id_list=id_list,
                                        team_id=self.uid,
                                        session=self.session,
                                        dataset_id=None,
                                        timeout=timeout,
                                        polling_delay=polling_delay)


class TeamCollection(AdminCollection[Team]):
    """
    Represents the collection of all teams as well as the resources belonging to it.

    Parameters
    ----------
    session: Session, optional
        The Citrine session used to connect to the database.

    """

    _path_template = '/teams'
    _individual_key = 'team'
    _collection_key = 'teams'
    _resource = Team
    _api_version = "v3"

    def __init__(self, session: Session):
        self.session = session

    def build(self, data) -> Team:
        """
        Build an individual team from a dictionary.

        Parameters
        ----------
        data: dict
            A dictionary representing the team.

        Return
        -------
        Team
            The team created from data.

        """
        team = Team.build(data)
        team.session = self.session
        return team

    def update(self, team: Team) -> Team:
        """
        Update a particular team.

        You can only update the name and/or description.
        Parameters
        ----------
        team: Team
            The Team Resource to be updated

        """
        url = self._get_path(team.uid)
        updated = self.session.patch_resource(url, team.dump(), version=self._api_version)
        data = updated[self._individual_key]
        return self.build(data)

    def register(self, name: str, *, description: str = "") -> Team:
        """
        Create and upload new team.

        Automatically grants user READ, WRITE, and SHARE access to this team.

        Parameters
        ----------
        name: str
            Name of the team to be created.
        description: str
            Long-form description of the team to be created.

        """
        team = super().register(Team(name, description=description))
        user_id = UserCollection(self.session).me().uid
        team.update_user_action(user_id=user_id, actions=[READ, WRITE, SHARE])
        return team



import functools
from typing import Iterator, Optional, Union
from uuid import UUID

from citrine.informatics.workflows.analysis_workflow import AnalysisWorkflow, \
    AnalysisWorkflowUpdatePayload
from citrine._rest.collection import Collection
from citrine._session import Session


class AnalysisWorkflowCollection(Collection[AnalysisWorkflow]):
    """Represents the collection of all analysis workflows for a team.

    Parameters
    ----------
    team_id: UUID
        the UUID of the team

    """

    _api_version = 'v1'
    _path_template = '/teams/{team_id}/analysis-workflows'
    _individual_key = None
    _resource = AnalysisWorkflow
    _collection_key = 'response'

    def __init__(self, session: Session, *, team_id: UUID):
        self.session = session
        self.team_id = team_id

    def build(self, data: dict) -> AnalysisWorkflow:
        """
        Build an individual analysis workflow from a dictionary.

        Parameters
        ----------
        data: dict
            A dictionary representing the analysis workflow.

        Return
        -------
        AnalysisWorkflow
            The analysis workflow created from data.

        """
        workflow = AnalysisWorkflow.build(data)
        workflow.team_id = self.team_id
        return workflow

    def list_all(self, *, per_page: int = 20) -> Iterator[AnalysisWorkflow]:
        """List all analysis workflows."""
        return self._list_with_params(per_page=per_page, include_archived=True)

    def list_archived(self, *, per_page: int = 20) -> Iterator[AnalysisWorkflow]:
        """List archived analysis workflows."""
        return self._list_with_params(per_page=per_page, filter="archived eq 'true'")

    def list(self, *, per_page: int = 20) -> Iterator[AnalysisWorkflow]:
        """List acttive analysis workflows."""
        return self._list_with_params(per_page=per_page, filter="archived eq 'false'")

    def _list_with_params(self, *, per_page: int, **kwargs) -> Iterator[AnalysisWorkflow]:
        page_fetcher = functools.partial(self._fetch_page, additional_params=kwargs)
        return self._paginator.paginate(page_fetcher=page_fetcher,
                                        collection_builder=self._build_collection_elements,
                                        per_page=per_page)

    def archive(self, uid: Union[UUID, str]) -> AnalysisWorkflow:
        """Archive an analysis workflow, hiding it from default listings."""
        url = self._get_path(uid=uid, action="archive")
        entity = self.session.put_resource(url, {}, version=self._api_version)
        return self.build(entity)

    def restore(self, uid: Union[UUID, str]) -> AnalysisWorkflow:
        """Restore an analysis workflow, showing it in default listings."""
        url = self._get_path(uid=uid, action="restore")
        entity = self.session.put_resource(url, {}, version=self._api_version)
        return self.build(entity)

    def update(self,
               uid: Union[UUID, str],
               *,
               name: Optional[str] = None,
               description: Optional[str] = None) -> AnalysisWorkflow:
        """Update the name and/or description of the analysis workflow."""
        aw_update = AnalysisWorkflowUpdatePayload(uid=uid, name=name, description=description)
        return super().update(aw_update)

    def rebuild(self, uid: Union[UUID, str]) -> AnalysisWorkflow:
        """Rebuild the data source underlying the analysis workflow."""
        url = self._get_path(uid=uid, action=("query", "rerun"))
        entity = self.session.put_resource(url, {}, version=self._api_version)
        return self.build(entity)

    def delete(self, uid: Union[UUID, str]):
        """Analysis workflows cannot be deleted at this time."""
        raise NotImplementedError("Deleting Analysis Workflows is not currently supported.")



"""Tools for working with Descriptors."""
from abc import abstractmethod
from typing import Type, List, Mapping, Optional, Union
from uuid import UUID
from warnings import warn

from citrine._serialization import properties
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization.serializable import Serializable
from citrine.informatics.descriptors import Descriptor
from citrine.resources.file_link import FileLink
from citrine.resources.gemtables import GemTable

__all__ = [
    'DataSource',
    'CSVDataSource',
    'GemTableDataSource',
    'ExperimentDataSourceRef',
    'SnapshotDataSource',
]


class DataSource(PolymorphicSerializable['DataSource']):
    """A source of data for the AI engine.

    Data source provides a polymorphic interface for specifying different kinds of data as the
    training data for predictors and the input data for some design spaces.

    """

    def __eq__(self, other):
        if isinstance(other, Serializable):
            return self.dump() == other.dump()
        else:
            return False

    @classmethod
    def _subclass_list(self) -> List[Type[Serializable]]:
        return [CSVDataSource, GemTableDataSource, ExperimentDataSourceRef, SnapshotDataSource]

    @classmethod
    def get_type(cls, data) -> Type[Serializable]:
        """Return the subtype."""
        if "type" not in data:
            raise ValueError("Can only get types from dicts with a 'type' key")
        res = next((x for x in cls._subclass_list() if x.typ == data["type"]), None)
        if res is None:
            raise ValueError(f"Unrecognized type: {data['type']}")
        return res

    @property
    @abstractmethod
    def _data_source_type(self) -> str:
        """The data source type string, which is the leading term of the data_source_id."""

    @classmethod
    def from_data_source_id(cls, data_source_id: str) -> "DataSource":
        """Build a DataSource from a datasource_id."""
        terms = data_source_id.split("::")
        res = next((x for x in cls._subclass_list() if x._data_source_type == terms[0]), None)
        if res is None:
            raise ValueError(f"Unrecognized type: {terms[0]}")
        return res._data_source_id_builder(*terms[1:])

    @classmethod
    @abstractmethod
    def _data_source_id_builder(cls, *args) -> "DataSource":
        """Build a DataSource based on a parsed data_source_id."""

    @abstractmethod
    def to_data_source_id(self) -> str:
        """Generate the data_source_id for this DataSource."""


class CSVDataSource(Serializable['CSVDataSource'], DataSource):
    """A data source based on a CSV file stored on the data platform.

    Parameters
    ----------
    file_link: FileLink
        link to the CSV file to read the data from
    column_definitions: Mapping[str, Descriptor]
        Map the column headers to the descriptors that will be used to interpret the cell contents
    identifiers: Optional[List[str]]
        List of one or more column headers whose values uniquely identify a row. These may overlap
        with ``column_definitions`` if a column should be used as data and as an identifier,
        but this is not necessary. Identifiers must be unique within a dataset. No two rows can
        contain the same value.

    """

    typ = properties.String('type', default='csv_data_source', deserializable=False)
    file_link = properties.Object(FileLink, "file_link")
    column_definitions = properties.Mapping(
        properties.String, properties.Object(Descriptor), "column_definitions")
    identifiers = properties.Optional(properties.List(properties.String), "identifiers")

    _data_source_type = "csv"

    def __init__(self,
                 *,
                 file_link: FileLink,
                 column_definitions: Mapping[str, Descriptor],
                 identifiers: Optional[List[str]] = None):
        self.file_link = file_link
        self.column_definitions = column_definitions
        self.identifiers = identifiers

    @classmethod
    def _data_source_id_builder(cls, *args) -> DataSource:
        # TODO Figure out how to populate the column definitions
        warn("A CSVDataSource was derived from a data_source_id "
             "but is missing its column_definitions and identities",
             UserWarning)
        return CSVDataSource(
            file_link=FileLink(url=args[0], filename=args[1]),
            column_definitions={}
        )

    def to_data_source_id(self) -> str:
        """Generate the data_source_id for this DataSource."""
        return f"{self._data_source_type}::{self.file_link.url}::{self.file_link.filename}"


class GemTableDataSource(Serializable['GemTableDataSource'], DataSource):
    """A data source based on a GEM Table hosted on the data platform.

    Parameters
    ----------
    table_id: UUID
        Unique identifier for the GEM Table
    table_version: Union[str,int]
        Version number for the GEM Table. The first GEM table built from a configuration
        has version = 1. Strings are cast to ints.

    """

    typ = properties.String('type', default='hosted_table_data_source', deserializable=False)
    table_id = properties.UUID("table_id")
    table_version = properties.Integer("table_version")

    _data_source_type = "gemd"

    def __init__(self,
                 *,
                 table_id: UUID,
                 table_version: Union[int, str]):
        self.table_id: UUID = table_id
        self.table_version: Union[int, str] = table_version

    @classmethod
    def _data_source_id_builder(cls, *args) -> DataSource:
        return GemTableDataSource(table_id=UUID(args[0]), table_version=args[1])

    def to_data_source_id(self) -> str:
        """Generate the data_source_id for this DataSource."""
        return f"{self._data_source_type}::{self.table_id}::{self.table_version}"

    @classmethod
    def from_gemtable(cls, table: GemTable) -> "GemTableDataSource":
        """Generate a DataSource that corresponds to a GemTable.

        Parameters
        ----------
        table: GemTable
            The GemTable object to reference

        """
        return GemTableDataSource(table_id=table.uid, table_version=table.version)


class ExperimentDataSourceRef(Serializable['ExperimentDataSourceRef'], DataSource):
    """A reference to a data source based on an experiment result hosted on the data platform.

    Parameters
    ----------
    datasource_id: UUID
        Unique identifier for the Experiment Data Source

    """

    typ = properties.String('type', default='experiments_data_source', deserializable=False)
    datasource_id = properties.UUID("datasource_id")

    _data_source_type = "experiments"

    def __init__(self, *, datasource_id: UUID):
        self.datasource_id: UUID = datasource_id

    @classmethod
    def _data_source_id_builder(cls, *args) -> DataSource:
        return ExperimentDataSourceRef(datasource_id=UUID(args[0]))

    def to_data_source_id(self) -> str:
        """Generate the data_source_id for this DataSource."""
        return f"{self._data_source_type}::{self.datasource_id}"


class SnapshotDataSource(Serializable['SnapshotDataSource'], DataSource):
    """A reference to a data source based on a Snapshot on the data platform.

    Parameters
    ----------
    snapshot_id: UUID
        Unique identifier for the Snapshot Data Source

    """

    typ = properties.String('type', default='snapshot_data_source', deserializable=False)
    snapshot_id = properties.UUID("snapshot_id")

    _data_source_type = "snapshot"

    def __init__(self, *, snapshot_id: UUID):
        self.snapshot_id = snapshot_id

    @classmethod
    def _data_source_id_builder(cls, *args) -> DataSource:
        return SnapshotDataSource(snapshot_id=UUID(args[0]))

    def to_data_source_id(self) -> str:
        """Generate the data_source_id for this DataSource."""
        return f"{self._data_source_type}::{self.snapshot_id}"



"""Definitions for GemdQuery objects, and their sub-objects."""
from typing import List, Type

from citrine._serialization.serializable import Serializable
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization import properties

__all__ = ['AllRealFilter', 'AllIntegerFilter', 'NominalCategoricalFilter']


class PropertyFilterType(PolymorphicSerializable):
    """Abstract concept of a criteria to apply when searching for materials."""

    @classmethod
    def get_type(cls, data) -> Type[Serializable]:
        """Return the subtype."""
        classes: List[Type[PropertyFilterType]] = [
            NominalCategoricalFilter,
            AllRealFilter, AllIntegerFilter
        ]
        return {klass.typ: klass for klass in classes}[data['type']]


class AllRealFilter(Serializable['AllRealFilter'], PropertyFilterType):
    """
    Filter for any real value that fits certain constraints.

    Parameters
    ----------
    lower: float
        The lower bound on this filter range.
    upper: float
        The upper bound on this filter range.
    unit: str
        The units associated with the floating point values for this filter.

    """

    lower = properties.Float('lower')
    upper = properties.Float('upper')
    unit = properties.String('unit')
    typ = properties.String('type', default="all_real_filter", deserializable=False)


class AllIntegerFilter(Serializable['AllIntegerFilter'], PropertyFilterType):
    """
    Filter for any integer value that fits certain constraints.

    Parameters
    ----------
    lower: float
        The lower bound on this filter range.
    upper: float
        The upper bound on this filter range.
    inclusive: bool
        Whether the lower & upper bounds are included in the range.

    """

    lower = properties.Float('lower')
    upper = properties.Float('upper')
    inclusive = properties.Optional(properties.Boolean, 'inclusive', default=True)
    typ = properties.String('type', default="all_integer_filter", deserializable=False)


class NominalCategoricalFilter(Serializable['NominalCategoricalFilter'], PropertyFilterType):
    """
    Filter based upon a fixed list of Categorical Values.

    Parameters
    ----------
    categories: Set[str]
        Which categorical values match.

    """

    categories = properties.Set(properties.String, 'categories')
    typ = properties.String('type', default="nominal_categorical_filter", deserializable=False)



from typing import List

from citrine._rest.resource import Resource
from citrine._serialization import properties as _properties
from citrine.informatics.descriptors import Descriptor
from citrine.informatics.predictors import PredictorNode


class AttributeAccumulationPredictor(Resource["AttributeAccumulationPredictor"], PredictorNode):
    """A node to support propagating attributes through training.

    You should never have to add this node yourself: the backend should be able to automatically
    create it when necessary.
    """

    attributes = _properties.List(_properties.Object(Descriptor), 'attributes')
    sequential = _properties.Boolean('sequential')

    typ = _properties.String('type', default='AttributeAccumulation', deserializable=False)

    def __init__(self,
                 name: str,
                 *,
                 description: str,
                 attributes: List[Descriptor],
                 sequential: bool):
        self.name = name
        self.description = description
        self.attributes = attributes
        self.sequential = sequential

    def __str__(self):
        return '<AttributeAccumulationPredictor {!r}>'.format(self.name)



"""Definitions for GemdQuery objects, and their sub-objects."""
from typing import List, Type

from gemd.enumeration.base_enumeration import BaseEnumeration

from citrine._serialization.serializable import Serializable
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization import properties
from citrine.gemd_queries.filter import PropertyFilterType

__all__ = ['MaterialClassification', 'TextSearchType',
           'AndOperator', 'OrOperator',
           'PropertiesCriteria', 'NameCriteria',
           'MaterialRunClassificationCriteria', 'MaterialTemplatesCriteria'
           ]


class MaterialClassification(BaseEnumeration):
    """A classification based on where in a Material History you find a Material."""

    ATOMIC_INGREDIENT = "atomic_ingredient"
    INTERMEDIATE_INGREDIENT = "intermediate_ingredient"
    TERMINAL_MATERIAL = "terminal_material"


class TextSearchType(BaseEnumeration):
    """The style of text search to run."""

    EXACT = "exact"
    PREFIX = "prefix"
    SUFFIX = "suffix"
    SUBSTRING = "substring"


class Criteria(PolymorphicSerializable):
    """Abstract concept of a criteria to apply when searching for materials."""

    @classmethod
    def get_type(cls, data) -> Type[Serializable]:
        """Return the subtype."""
        classes: List[Type[Criteria]] = [
            AndOperator, OrOperator,
            PropertiesCriteria, NameCriteria, MaterialRunClassificationCriteria,
            MaterialTemplatesCriteria
        ]
        return {klass.typ: klass for klass in classes}[data['type']]


class AndOperator(Serializable['AndOperator'], Criteria):
    """
    Combine multiple criteria, requiring EACH to be true for a match.

    Parameters
    ----------
    criteria: Criteria
        List of conditions all responses must satisfy (i.e., joined with an AND).

    """

    criteria = properties.List(properties.Object(Criteria), "criteria")
    typ = properties.String('type', default="and_operator", deserializable=False)


class OrOperator(Serializable['OrOperator'], Criteria):
    """
    Combine multiple criteria, requiring ANY to be true for a match.

    Parameters
    ----------
    criteria: Criteria
        List of conditions, at least one of which must match (i.e., joined with an OR).

    """

    criteria = properties.List(properties.Object(Criteria), "criteria")
    typ = properties.String('type', default="or_operator", deserializable=False)


class PropertiesCriteria(Serializable['PropertiesCriteria'], Criteria):
    """
    Look for materials with a particular Property and optionally Value types & ranges.

    Parameters
    ----------
    property_templates_filter: Set[UUID]
        The citrine IDs of the property templates matches must reference.
    value_type_filter: Optional[PropertyFilterType]
        The value range matches must conform to.

    """

    property_templates_filter = properties.Set(properties.UUID, "property_templates_filter")
    value_type_filter = properties.Optional(
        properties.Object(PropertyFilterType), "value_type_filter"
    )
    typ = properties.String('type', default="properties_criteria", deserializable=False)


class NameCriteria(Serializable['NameCriteria'], Criteria):
    """
    Look for materials with particular names.

    Parameters
    ----------
    name: str
        The name the returned objects must have.
    search_type: TextSearchType
        What kind of string match to use (exact, substring, ...).

    """

    name = properties.String('name')
    search_type = properties.Enumeration(TextSearchType, 'search_type')
    typ = properties.String('type', default="name_criteria", deserializable=False)


class MaterialRunClassificationCriteria(
    Serializable['MaterialRunClassificationCriteria'],
    Criteria
):
    """
    Look for materials with particular classification, defined by MaterialClassification.

    Parameters
    ----------
    classifications: Set[MaterialClassification]
        The classification, based on where in a material history an object appears.

    """

    classifications = properties.Set(
        properties.Enumeration(MaterialClassification), 'classifications'
    )
    typ = properties.String(
        'type',
        default="material_run_classification_criteria",
        deserializable=False
    )


class MaterialTemplatesCriteria(Serializable['MaterialTemplatesCriteria'], Criteria):
    """
    Look for materials with particular Material Templates and tags.

    This has a similar behavior to the old [[MaterialRunByTemplate]] Row definition

    Parameters
    ----------
    material_templates_identifiers: Set[UUID]
        Which material templates to filter by.
    tag_filters: Set[str]
        Which tags to filter by.

    """

    material_templates_identifiers = properties.Set(
        properties.UUID,
        "material_templates_identifiers"
    )
    tag_filters = properties.Set(properties.String, 'tag_filters')
    typ = properties.String('type', default="material_template_criteria", deserializable=False)



from typing import Dict
from uuid import UUID

from citrine._rest.resource import Resource
from citrine._serialization import properties


class ShapleyMaterial(Resource):
    """The feature effect of a material."""

    material_id = properties.UUID('material_id', serializable=False)
    value = properties.Float('value', serializable=False)


class ShapleyFeature(Resource):
    """All feature effects for this feature by material."""

    feature = properties.String('feature', serializable=False)
    materials = properties.List(properties.Object(ShapleyMaterial), 'materials',
                                serializable=False)

    @property
    def material_dict(self) -> Dict[UUID, float]:
        """Presents the feature's effects as a dictionary by material."""
        return {material.material_id: material.value for material in self.materials}


class ShapleyOutput(Resource):
    """All feature effects for this output by feature."""

    output = properties.String('output', serializable=False)
    features = properties.List(properties.Object(ShapleyFeature), 'features', serializable=False)

    @property
    def feature_dict(self) -> Dict[str, Dict[UUID, float]]:
        """Presents the output's feature effects as a dictionary by feature."""
        return {feature.feature: feature.material_dict for feature in self.features}


class FeatureEffects(Resource):
    """Captures information about the feature effects associated with a predictor."""

    predictor_id = properties.UUID('metadata.predictor_id', serializable=False)
    predictor_version = properties.Integer('metadata.predictor_version', serializable=False)
    status = properties.String('metadata.status', serializable=False)
    failure_reason = properties.Optional(properties.String(), 'metadata.failure_reason',
                                                              serializable=False)

    outputs = properties.Optional(properties.List(properties.Object(ShapleyOutput)), 'resultobj',
                                  serializable=False)

    @classmethod
    def _pre_build(cls, data: dict) -> Dict:
        shapley = data.get("result")
        if not shapley:
            return data

        material_ids = shapley["materials"]

        outputs = []
        for output, feature_dict in shapley["outputs"].items():
            features = []
            for feature, values in feature_dict.items():
                items = zip(material_ids, values)
                materials = [{"material_id": mid, "value": value} for mid, value in items]
                features.append({
                    "feature": feature,
                    "materials": materials
                })

            outputs.append({"output": output, "features": features})

        data["resultobj"] = outputs
        return data

    @property
    def as_dict(self) -> Dict[str, Dict[str, Dict[UUID, float]]]:
        """Presents the feature effects as a dictionary by output."""
        if self.outputs:
            return {output.output: output.feature_dict for output in self.outputs}
        else:
            return {}



from typing import List, Optional

from citrine.resources.data_concepts import CITRINE_TAG_PREFIX

_CITRINE_DEFAULT_LABEL_PREFIX = f'{CITRINE_TAG_PREFIX}::mat_label'


def _inject_default_label_tags(
    original_tags: Optional[List[str]], default_labels: Optional[List[str]]
) -> Optional[List[str]]:
    if default_labels is None:
        all_tags = original_tags
    else:
        labels_as_tags = [
            f"{_CITRINE_DEFAULT_LABEL_PREFIX}::{label}" for label in default_labels
        ]
        if original_tags is None:
            all_tags = labels_as_tags
        else:
            all_tags = list(original_tags) + labels_as_tags
    return all_tags



"""Definitions for GemdQuery objects, and their sub-objects."""
from gemd.enumeration.base_enumeration import BaseEnumeration

from citrine._serialization.serializable import Serializable
from citrine._serialization import properties
from citrine.gemd_queries.criteria import Criteria


class GemdObjectType(BaseEnumeration):
    """The style of text search to run."""

    # An old defect has some old GemdQuery values stored with invalid enums
    # The synonyms will allow invalid old values to be read, but not emitted
    MEASUREMENT_TEMPLATE_TYPE = "measurement_template", "MEASUREMENT_TEMPLATE_TYPE"
    MATERIAL_TEMPLATE_TYPE = "material_template", "MATERIAL_TEMPLATE_TYPE"
    PROCESS_TEMPLATE_TYPE = "process_template", "PROCESS_TEMPLATE_TYPE"
    PROPERTY_TEMPLATE_TYPE = "property_template", "PROPERTY_TEMPLATE_TYPE"
    CONDITION_TEMPLATE_TYPE = "condition_template", "CONDITION_TEMPLATE_TYPE"
    PARAMETER_TEMPLATE_TYPE = "parameter_template", "PARAMETER_TEMPLATE_TYPE"
    PROCESS_RUN_TYPE = "process_run", "PROCESS_RUN_TYPE"
    PROCESS_SPEC_TYPE = "process_spec", "PROCESS_SPEC_TYPE"
    MATERIAL_RUN_TYPE = "material_run", "MATERIAL_RUN_TYPE"
    MATERIAL_SPEC_TYPE = "material_spec", "MATERIAL_SPEC_TYPE"
    INGREDIENT_RUN_TYPE = "ingredient_run", "INGREDIENT_RUN_TYPE"
    INGREDIENT_SPEC_TYPE = "ingredient_spec", "INGREDIENT_SPEC_TYPE"
    MEASUREMENT_RUN_TYPE = "measurement_run", "MEASUREMENT_RUN_TYPE"
    MEASUREMENT_SPEC_TYPE = "measurement_spec", "MEASUREMENT_SPEC_TYPE"


class GemdQuery(Serializable['GemdQuery']):
    """
    This describes what data objects to fetch (or graph of data objects).

    Parameters
    ----------
    criteria: Criteria
        List of conditions all responses must satisfy (i.e., joined with an AND).
    datasets: UUID
        Set of datasets to look in for matching objects.
    object_types: GemdObjectType
        Classes of objects to consider when searching.
    schema_version: Int
        What version of the query schema this package represents.

    """

    criteria = properties.List(properties.Object(Criteria), "criteria", default=[])
    datasets = properties.Set(properties.UUID, "datasets", default=set())
    object_types = properties.Set(
        properties.Enumeration(GemdObjectType),
        'object_types',
        default={x for x in GemdObjectType}
    )
    schema_version = properties.Integer('schema_version', default=1)

    @classmethod
    def _pre_build(cls, data: dict) -> dict:
        """Run data modification before building."""
        version = data.get('schema_version')
        if data.get('schema_version') != 1:
            raise ValueError(
                f"This version of the library only supports schema_version 1, not '{version}'"
            )
        return data



import functools
from typing import Iterator, Optional, Union
from uuid import UUID

from citrine._rest.collection import Collection
from citrine._rest.resource import Resource
from citrine._serialization import properties
from citrine._session import Session
from citrine.exceptions import NotFound
from citrine.resources.data_version_update import BranchDataUpdate, NextBranchVersionRequest
from citrine.resources.design_workflow import DesignWorkflowCollection
from citrine.resources.experiment_datasource import (ExperimentDataSourceCollection,
                                                     ExperimentDataSource)


LATEST_VER = "latest"  # Refers to the most recently created branch version.


class Branch(Resource['Branch']):
    """
    A project branch.

    A branch is a container for design workflows.
    """

    name = properties.String('data.name')
    uid = properties.Optional(properties.UUID(), 'id')
    archived = properties.Boolean('metadata.archived', serializable=False)
    created_at = properties.Optional(properties.Datetime(), 'metadata.created.time',
                                     serializable=False)
    updated_at = properties.Optional(properties.Datetime(), 'metadata.updated.time',
                                     serializable=False)
    # added in v2
    root_id = properties.UUID('metadata.root_id', serializable=False)
    version = properties.Integer('metadata.version', serializable=False)

    project_id: Optional[UUID] = None

    def __init__(self,
                 name: str,
                 *,
                 session: Optional[Session] = None):
        self.name: str = name
        self.session: Session = session

    def __str__(self):
        return f'<Branch {self.name!r}>'

    @property
    def design_workflows(self) -> DesignWorkflowCollection:
        """Return a resource representing all workflows contained within this branch."""
        if getattr(self, 'project_id', None) is None:
            raise AttributeError('Cannot initialize workflow without project reference!')
        return DesignWorkflowCollection(project_id=self.project_id,
                                        session=self.session,
                                        branch_root_id=self.root_id,
                                        branch_version=self.version)

    @property
    def experiment_datasource(self) -> Optional[ExperimentDataSource]:
        """Return this branch's experiment data source, or None if one doesn't exist."""
        if getattr(self, 'project_id', None) is None:
            raise AttributeError('Cannot retrieve datasource without project reference!')
        erds = ExperimentDataSourceCollection(project_id=self.project_id, session=self.session)
        branch_erds_iter = erds.list(branch_version_id=self.uid, version=LATEST_VER)
        return next(branch_erds_iter, None)

    def _post_dump(self, data: dict) -> dict:
        # Only the data portion of an entity is sent to the server.
        data = data["data"]
        return data


class BranchCollection(Collection[Branch]):
    """A collection of Branches."""

    _path_template = '/projects/{project_id}/branches'
    _individual_key = None
    _collection_key = 'response'
    _resource = Branch
    _api_version = 'v2'

    def __init__(self, project_id: UUID, session: Session):
        self.project_id: UUID = project_id
        self.session: Session = session

    def build(self, data: dict) -> Branch:
        """
        Build an individual branch from a dictionary.

        Parameters
        ----------
        data: dict
            A dictionary representing the branch.

        Return
        -------
        Branch
            The branch created from data.

        """
        branch = Branch.build(data)
        branch.session = self.session
        branch.project_id = self.project_id
        return branch

    def get(self,
            *,
            root_id: Union[UUID, str],
            version: Optional[Union[int, str]] = LATEST_VER) -> Branch:
        """
        Retrieve a branch by its root ID and, optionally, its version number.

        Omitting the version number will retrieve the latest version.

        Parameters
        ---------
        root_id: Union[UUID, str]
            Unique identifier of the branch root

        version: Union[int, str], optional
            The version of the branch to retrieve. If provided, must either be a positive integer,
            or "latest". Defaults to "latest".

        Returns
        -------
        Branch
            The requested version of the branch.

        """
        version = version or LATEST_VER
        params = {"root": str(root_id), "version": version}
        branch = next(self._list_with_params(per_page=1, **params), None)
        if branch:
            return branch
        else:
            raise NotFound.build(
                message=f"Branch root '{root_id}', version {version} not found",
                method="GET",
                path=self._get_path(),
                params=params
            )

    def get_by_version_id(self, *, version_id: Union[UUID, str]) -> Branch:
        """
        Given a branch version ID, retrieve the branch.

        Parameters
        ---------
        version_id: Union[UUID, str]
            Unique ID of the branch version

        Returns
        -------
        Branch
            The requested branch.

        """
        return super().get(version_id)

    def list(self, *, per_page: int = 20) -> Iterator[Branch]:
        """
        List all active branches using pagination.

        Yields all active branches paginating over all available pages.

        Parameters
        ---------
        per_page: int, optional
            Max number of branches to return per page. Default is 20. This parameter is used when
            making requests to the backend service.

        Returns
        -------
        Iterator[Branch]
            All branches in this collection.

        """
        return self._list_with_params(per_page=per_page, archived=False)

    def list_archived(self, *, per_page: int = 20) -> Iterator[Branch]:
        """
        List archived branches using pagination.

        Yields all archived branches, paginating over all available pages.

        Parameters
        ---------
        per_page: int, optional
            Max number of branches to return per page. Default is 20. This parameter is used when
            making requests to the backend service.

        Returns
        -------
        Iterator[Branch]
            Archived branches in this collection.

        """
        return self._list_with_params(per_page=per_page, archived=True)

    def list_all(self, *, per_page: int = 20) -> Iterator[Branch]:
        """
        List active branches using pagination.

        Yields all active branches, paginating over all available pages.

        Parameters
        ---------
        per_page: int, optional
            Max number of branches to return per page. Default is 20. This parameter is used when
            making requests to the backend service.

        Returns
        -------
        Iterator[Branch]
            Active branches in this collection.

        """
        return self._list_with_params(per_page=per_page)

    def _list_with_params(self, *, per_page, **kwargs):
        fetcher = functools.partial(self._fetch_page, additional_params=kwargs)
        return self._paginator.paginate(page_fetcher=fetcher,
                                        collection_builder=self._build_collection_elements,
                                        per_page=per_page)

    def archive(self,
                *,
                root_id: Union[UUID, str],
                version: Optional[Union[int, str]] = LATEST_VER):
        """
        Archive a branch.

        Parameters
        ----------
        root_id: Union[UUID, str]
            Unique ID of the branch root

        version: Union[int, str], optional
            The version of the branch. If provided, must either be a positive integer, or "latest".
            Defaults to "latest".

        """
        version = version or LATEST_VER
        # The backend API at present expects a branch version ID, so we must look it up.
        uid = self.get(root_id=root_id, version=version).uid

        url = self._get_path(uid, action="archive")
        data = self.session.put_resource(url, {}, version=self._api_version)
        return self.build(data)

    def restore(self,
                *,
                root_id: Union[UUID, str],
                version: Optional[Union[int, str]] = LATEST_VER):
        """
        Restore an archived branch.

        Parameters
        ----------
        root_id: Union[UUID, str]
            Unique ID of the branch root

        version: Union[int, str], optional
            The version of the branch. If provided, must either be a positive integer, or "latest".
            Defaults to "latest".

        """
        version = version or LATEST_VER
        # The backend API at present expects a branch version ID, so we must look it up.
        uid = self.get(root_id=root_id, version=version).uid

        url = self._get_path(uid, action="restore")
        data = self.session.put_resource(url, {}, version=self._api_version)
        return self.build(data)

    def update_data(self,
                    *,
                    root_id: Union[UUID, str],
                    version: Optional[Union[int, str]] = LATEST_VER,
                    use_existing: bool = True,
                    retrain_models: bool = False) -> Optional[Branch]:
        """
        Automatically advance the branch to the next version.

        If there are no newer versions of data sources used by this branch this method returns
        without doing anything

        Parameters
        ----------
        root_id: Union[UUID, str]
            Unique ID of the branch root

        version: Union[int, str], optional
            The version of the branch. If provided, must either be a positive integer, or "latest".
            Defaults to "latest".

        use_existing: bool
            If true the workflows in this branch will use existing predictors that are using
            the latest versions of the data sources and are ready to use.

        retrain_models: bool
            If true, when new versions of models are created, they are automatically
            scheduled for training.

        Returns
        -------
        Branch
            The new branch record after version update or None if no update

        """
        version = version or LATEST_VER
        version_updates = self.data_updates(root_id=root_id, version=version)
        # If no new data sources, then exit, nothing to do
        if len(version_updates.data_updates) == 0:
            return None

        use_predictors = []
        if use_existing:
            use_predictors = version_updates.predictors

        branch_instructions = NextBranchVersionRequest(data_updates=version_updates.data_updates,
                                                       use_predictors=use_predictors)
        branch = self.next_version(root_id=root_id,
                                   branch_instructions=branch_instructions,
                                   retrain_models=retrain_models)
        return branch

    def data_updates(self,
                     *,
                     root_id: Union[UUID, str],
                     version: Optional[Union[int, str]] = LATEST_VER) -> BranchDataUpdate:
        """
        Get data updates for a branch.

        Parameters
        ----------
        root_id: Union[UUID, str]
            Unique ID of the branch root

        version: Union[int, str], optional
            The version of the branch. If provided, must either be a positive integer, or "latest".
            Defaults to "latest".

        Returns
        -------
        BranchDataUpdate
            A list of data updates and compatible predictors

        """
        version = version or LATEST_VER
        # The backend API at present expects a branch version ID, so we must look it up.
        uid = self.get(root_id=root_id, version=version).uid

        path = self._get_path(uid=uid, action="data-version-updates-predictor")
        data = self.session.get_resource(path, version=self._api_version)
        return BranchDataUpdate.build(data)

    def next_version(self,
                     root_id: Union[UUID, str],
                     *,
                     branch_instructions: NextBranchVersionRequest,
                     retrain_models: bool = True):
        """
        Move a branch to the next version.

        Parameters
        ----------
        root_id: Union[UUID, str]
            Unique identifier of the branch root to advance to next version

        branch_instructions: NextBranchVersionRequest
            Instructions for how the next version of a branch should handle its predictors
            when the workflows are cloned.  data_updates contains the list of data source
            versions to upgrade (current->latest), and use_predictors will either have a
            <predictor_id>:latest to indicate the workflow should use a new version of
            the predictor.  Or <predictor_id>:<version #> to indicate that the workflow
            should use an existing predictor version.

        retrain_models: bool
            If true, when new versions of models are created, they are automatically
            scheduled for training

        Returns
        -------
        Branch
            The new branch record after version update

        """
        path = self._get_path(action="next-version-predictor")
        data = self.session.post_resource(path, branch_instructions.dump(),
                                          version=self._api_version,
                                          params={
                                              'root': str(root_id),
                                              'retrain_models': retrain_models})
        return self.build(data)
